@@ -36,13 +36,13 @@ use eyre::{Result, WrapErr};
 use fst::{Automaton, IntoStreamer, Map, Streamer};
 use std::{
     cmp::Reverse,
-    collections::{HashMap, HashSet},
+    collections::HashSet,
     fs::File,
     io::{BufRead, BufReader},
     path::{Path, PathBuf},
     str::FromStr,
 };
-use wordle_automaton::{Letter, Wordle, WordleBuilder};
+use wordle_automaton::{Guess, Letter, Wordle, WordleBuilder};
 
 #[cfg(feature = "generate")]
 mod words;
@@ -94,19 +94,29 @@ fn main() -> Result<()> {
 }
 
 fn run<const N: usize>(opts: &Opts) -> Result<()> {
-    let words = load_word_list::<N>(
-        &opts.word_list,
-        opts.block_list.as_deref(),
-        opts.sort,
-        opts.penalty.0,
-    )?;
+    let words = load_word_list::<N>(&opts.word_list, opts.block_list.as_deref(), opts.sort)?;
     let fsts = build_fst(words)?;
     loop {
         let solution = solve::<N>(&fsts)?;
         match solution {
-            Solution::Solved(mut solution) => {
+            Solution::Solved(mut solution, guesses) => {
                 solution.make_ascii_uppercase();
-                println!("Solution: {}", solution);
+                println!("Solution: {solution}");
+                println!();
+                println!("Wordle {}/6", guesses.len());
+                for guess in guesses {
+                    let guess = guess
+                        .iter()
+                        .map(|g| match g {
+                            Guess::Absent => '⬛',
+                            Guess::Present => '🟨',
+                            Guess::Correct => '🟩',
+                        })
+                        .collect::<String>();
+                    println!("{guess}");
+                }
+                println!();
+                println!("Solution: {solution}");
             }
             Solution::None => println!("Could not find a solution"),
             Solution::Restart => continue,
@@ -123,29 +133,11 @@ struct Opts {
     word_list: PathBuf,
     block_list: Option<PathBuf>,
     sort: bool,
-    penalty: Penalty,
     size: Size,
     #[cfg(feature = "generate")]
     generate: bool,
 }
 
-#[derive(Debug)]
-struct Penalty(f64);
-
-impl FromStr for Penalty {
-    type Err = String;
-
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        s.parse::<f64>().map_err(|e| e.to_string()).and_then(|f| {
-            if (0.0..=1.0).contains(&f) {
-                Ok(Self(1.0 - f))
-            } else {
-                Err(String::from("The value must be in [0..1]"))
-            }
-        })
-    }
-}
-
 #[derive(Debug)]
 struct Size(usize);
 
@@ -206,25 +198,6 @@ fn parse_opts() -> Opts {
                 .short('c')
                 .long("sort"),
         )
-        .arg(
-            Arg::new("penalty")
-                .help("Penalty for words with duplicated letters")
-                .validator(str::parse::<Penalty>)
-                .long_help(concat!(
-                    "Penalty for a waord having duplicated letters. ",
-                    "The default behavior is prioritize words that have no repeated characters and ",
-                    "only onclude the other words if their letters are quite frequent. ",
-                    "This flag controls how much penalty a score gets. ",
-                    "The penalty must be between 0 and 1, inclusive.",
-                    "A penalty of 0 means no penalty, score the same as other words.",
-                    "A penalty of 1 means to only show words with duplicates after all other results.",
-                ))
-                .short('p')
-                .long("penalty")
-                .takes_value(true)
-                .required(false)
-                .default_value("0.5"),
-        )
         .arg(
             Arg::new("size")
                 .help("word size")
@@ -249,14 +222,12 @@ fn parse_opts() -> Opts {
     let word_list = PathBuf::from(matches.value_of_os("word-list").unwrap());
     let block_list = matches.value_of_os("block-list").map(PathBuf::from);
     let sort = matches.is_present("sort");
-    let penalty = matches.value_of_t("penalty").unwrap();
     let size = matches.value_of_t("size").unwrap();
 
     Opts {
         word_list,
         block_list,
         sort,
-        penalty,
         size,
         #[cfg(feature = "generate")]
         generate: matches.is_present("generate"),
@@ -267,7 +238,6 @@ fn load_word_list<const N: usize>(
     file: &Path,
     block_list: Option<&Path>,
     sort: bool,
-    penalty: f64,
 ) -> Result<Vec<(String, u64)>> {
     let lines = BufReader::new(
         File::open(file).wrap_err_with(|| format!("The file '{}' is missing.", file.display()))?,
@@ -290,7 +260,6 @@ fn load_word_list<const N: usize>(
         lines.lines().map_while(Result::ok),
         &block_list.unwrap_or_default(),
         sort,
-        penalty,
     ))
 }
 
@@ -298,7 +267,6 @@ fn clean_word_list<I, const N: usize>(
     words: I,
     block_list: &HashSet<String>,
     sort: bool,
-    penalty: f64,
 ) -> Vec<(String, u64)>
 where
     I: IntoIterator,
@@ -308,30 +276,22 @@ where
         (word.len() == N || word.len() == N - 1) && word.bytes().all(|b| matches!(b, b'a'..=b'z'))
     }
 
-    fn has_duplicate_letters(word: &str) -> bool {
-        let mut word = word.as_bytes();
-        while let Some((head, tail)) = word.split_first() {
-            if tail.contains(head) {
-                return true;
-            }
-            word = tail;
-        }
-        false
-    }
+    const EMPTY_FREQ: [u32; 26] = [0; 26];
 
     let mut possible_words = Vec::with_capacity(1024);
     let mut possible_stems = HashSet::with_capacity(1024);
     let mut possible_plurals = HashSet::with_capacity(1024);
-    let mut letter_frequency = HashMap::<u8, u64>::with_capacity(32);
+    let mut letter_frequency = EMPTY_FREQ;
 
     for word in words {
         let word_ref = word.as_ref();
 
-        // build frequency table
+        // build frequency table of all letters, ignoring their position
         if word_ref.is_ascii() {
             // maps uppercase to lowercase, ignore non alphabetic chars
             for byte in word_ref.bytes().filter_map(Letter::try_new).map(u8::from) {
-                *letter_frequency.entry(byte).or_default() += 1;
+                let idx = (byte - b'a') as usize;
+                letter_frequency[idx] += 1;
             }
         }
 
@@ -349,34 +309,33 @@ where
         }
     }
 
-    let is_singular = move |(idx, word): (usize, String)| {
-        (!(possible_plurals.contains(&idx) && possible_stems.contains(&word[..N - 1])))
-            .then(|| word)
-    };
+    let mut words = Vec::with_capacity(possible_words.len());
 
-    let score = move |s: &str| -> u64 {
-        s.bytes()
-            .map(|b| letter_frequency.get(&b).copied().unwrap_or_default())
-            .sum()
-    };
+    // build frequency table of letters at their respective position
+    let mut letter_pos_frequency = [EMPTY_FREQ; N];
 
-    let mut words = possible_words
+    for (idx, word) in possible_words.into_iter().enumerate() {
+        if possible_plurals.contains(&idx) && possible_stems.contains(&word[..N - 1]) {
+            continue;
+        }
+        for (pos, byte) in word.bytes().enumerate() {
+            let idx = (byte - b'a') as usize;
+            letter_pos_frequency[pos][idx] += 1;
+        }
+
+        words.push(word);
+    }
+
+    let mut words = words
         .into_iter()
-        .enumerate()
-        .filter_map(is_singular)
         .map(|word| {
-            let mut score = score(&word);
-            if has_duplicate_letters(&word) {
-                // score and penalty are both bound in a way that will not run into those issues
-                #[allow(
-                    clippy::cast_possible_truncation,
-                    clippy::cast_precision_loss,
-                    clippy::cast_sign_loss
-                )]
-                let new_score = ((score as f64) * penalty) as u64;
-                score = new_score;
+            let mut freq = [0; 26];
+            for (pos, b) in word.bytes().enumerate() {
+                let idx = (b - b'a') as usize;
+                // no adding, duplicate letters should not contribute multiple times
+                freq[idx] = letter_pos_frequency[pos][idx].saturating_add(letter_frequency[idx]);
             }
-            (word, score)
+            (word, freq.into_iter().map(u64::from).sum())
         })
         .collect::<Vec<_>>();
 
@@ -391,8 +350,8 @@ fn build_fst(words: Vec<(String, u64)>) -> Result<Map<Vec<u8>>> {
     Map::from_iter(words).wrap_err("The input file must be sorted. Try adding --sort.")
 }
 
-enum Solution {
-    Solved(String),
+enum Solution<const N: usize> {
+    Solved(String, Vec<[Guess; N]>),
     None,
     Restart,
     Quit,
@@ -400,39 +359,60 @@ enum Solution {
 
 const QUIT: &str = "-- QUIT I don't want to play anymore";
 
-fn solve<const N: usize>(fst: &Map<Vec<u8>>) -> Result<Solution> {
+fn solve<const N: usize>(fst: &Map<Vec<u8>>) -> Result<Solution<N>> {
     let mut wordle = WordleBuilder::<N>::new().build();
+    let mut guesses = Vec::with_capacity(6);
 
     let mut solutions = Vec::with_capacity(fst.len());
+    let wordle = loop {
+        let (next_worlde, word) = match guess_next(fst, wordle, &mut solutions)? {
+            Ok(next) => next,
+            Err(solution) => return Ok(solution),
+        };
 
-    loop {
-        if let Some(solution) = find_all_solutions(fst, &wordle, &mut solutions) {
-            return Ok(solution);
+        wordle = next_worlde;
+
+        guesses.push(extract_guess(&wordle, &word));
+        if wordle.is_solved() {
+            break wordle;
         }
+    };
 
-        eprintln!("Found {} possible solutions", solutions.len());
+    Ok(Solution::Solved(wordle.decode_str(), guesses))
+}
 
-        let word = match find_word_to_play(&wordle, &mut solutions)? {
-            Ok(word) => word,
-            Err(solution) => return Ok(solution),
-        };
-
-        wordle = match apply_feedback(wordle, word)? {
-            Ok(wordle) => wordle,
-            Err(solution) => return Ok(solution),
-        };
+fn guess_next<const N: usize>(
+    fst: &Map<Vec<u8>>,
+    wordle: Wordle<N>,
+    solutions: &mut Vec<(Vec<u8>, u64)>,
+) -> Result<Result<(Wordle<N>, Vec<u8>), Solution<N>>> {
+    if let Some(solution) = find_all_solutions(fst, &wordle, solutions) {
+        match solution {
+            Ok(word) => return Ok(Ok((mark_as_correct(wordle, &word), word))),
+            Err(solution) => return Ok(Err(solution)),
+        }
     }
+
+    eprintln!("Found {} possible solutions", solutions.len());
+
+    let word = match find_word_to_play(&wordle, solutions)? {
+        Ok(word) => word,
+        Err(solution) => return Ok(Err(solution)),
+    };
+
+    let wordle = match apply_feedback(wordle, &word)? {
+        Ok(wordle) => wordle,
+        Err(solution) => return Ok(Err(solution)),
+    };
+
+    Ok(Ok((wordle, word)))
 }
 
 fn find_all_solutions<const N: usize>(
     fst: &Map<Vec<u8>>,
     wordle: &Wordle<N>,
     buf: &mut Vec<(Vec<u8>, u64)>,
-) -> Option<Solution> {
-    if wordle.is_solved() {
-        return Some(Solution::Solved(wordle.decode_str()));
-    }
-
+) -> Option<Result<Vec<u8>, Solution<N>>> {
     buf.clear();
 
     let mut stream = fst.search(&wordle).into_stream();
@@ -441,13 +421,12 @@ fn find_all_solutions<const N: usize>(
     }
 
     if buf.is_empty() {
-        return Some(Solution::None);
+        return Some(Err(Solution::None));
     }
 
     if buf.len() == 1 {
         let (word, _) = buf.pop().unwrap();
-        let word = String::from_utf8(word).expect("input was strings");
-        return Some(Solution::Solved(word));
+        return Some(Ok(word));
     }
 
     buf.sort_by_key(|(_, score)| Reverse(*score));
@@ -455,10 +434,18 @@ fn find_all_solutions<const N: usize>(
     None
 }
 
+fn mark_as_correct<const N: usize>(wordle: Wordle<N>, word: &[u8]) -> Wordle<N> {
+    let mut wb = WordleBuilder::from(wordle);
+    word.iter()
+        .enumerate()
+        .fold(&mut wb, |wb, (pos, b)| wb.correct_pos(pos, *b))
+        .build()
+}
+
 fn find_word_to_play<const N: usize>(
     wordle: &Wordle<N>,
     options: &mut [(Vec<u8>, u64)],
-) -> Result<Result<Vec<u8>, Solution>> {
+) -> Result<Result<Vec<u8>, Solution<N>>> {
     let mut items = &mut *options;
 
     Ok(Ok(loop {
@@ -536,12 +523,12 @@ fn test_word<const N: usize>(word: &str, wordle: &Wordle<N>) -> bool {
 
 fn apply_feedback<const N: usize>(
     wordle: Wordle<N>,
-    word: Vec<u8>,
-) -> Result<Result<Wordle<N>, Solution>> {
+    word: &[u8],
+) -> Result<Result<Wordle<N>, Solution<N>>> {
     let mut wb = WordleBuilder::from(wordle);
     let mut selection = 0;
 
-    for (pos, b) in word.into_iter().enumerate() {
+    for (pos, &b) in word.iter().enumerate() {
         if wb.current().solution_at(pos) == Some(b) {
             eprintln!(
                 "Letter '{}' on position {}: Green:  Correct Letter in Correct Position",
@@ -570,7 +557,9 @@ fn apply_feedback<const N: usize>(
 
         let _ = match selection {
             0 => wb.never(b),
+            1 if wb.current().has_solution_at(pos) => wb.eventually(b),
             1 => wb.wrong_pos(pos, b),
+            2 if wb.current().has_solution_at(pos) => return Ok(Err(Solution::None)),
             2 => wb.correct_pos(pos, b),
             3 => wb.eventually(b),
             4 => &mut wb,
@@ -580,3 +569,12 @@ fn apply_feedback<const N: usize>(
 
     Ok(Ok(wb.build()))
 }
+
+fn extract_guess<const N: usize>(wordle: &Wordle<N>, word: &[u8]) -> [Guess; N] {
+    let mut pos = 0;
+    [(); N].map(|_| {
+        let guess = wordle.decode_guess(pos, word[pos]);
+        pos += 1;
+        guess
+    })
+}
